#include "dm.h"

#define NUM 4

dmFunctions dmf; // 描画関数の構造体
dsFunctions fn;

dWorldID world = NULL;                     // 動力学の世界
dSpaceID space = NULL;  // 衝突検出用スペース
dGeomID  ground = NULL; // 地面
dJointGroupID contactgroup = NULL; // コンタクトグループ

dmObject wheel[NUM], base, neck, camera, ball;
dJointID joint[NUM], neck_joint, camera_joint;
static const dReal  BALL_R    = 0.11;
static const dReal  BALL_P[3] = {1.0, 1.0, 0.14};
static dReal v_r = 0.0,v_l = 0.0 ;                   // 右,左車輪の回転速度
static dReal BASE_M    = 9.4;                        // 台車の質量
static dReal BASE_S[3] = {0.4,0.4,0.1};              // 台車のサイズ
static const dReal  CAMERA_M  = 0.2;
static const dReal  CAMERA_S[3]  = {0.2,0.2,0.05};
static const dReal  NECK_M    = 0.5;
static const dReal  NECK_L    = 0.2;
static const dReal  NECK_R    = 0.025;
static dReal WH_R0 = 0.05, WH_R1 = 0.101;            // 車輪0,1の半径
static dReal WH_W  = 0.02, WH_M0 = 0.1, WH_M1 = 0.2; // 車輪の幅,質量
static dReal START_X = 0, START_Y = 0, START_Z=0.20; // 初期位置

int CN1;

/*** 描画関数の設定 ***/
void setDrawStuff()
{
    fn.version = DS_VERSION;    // ドロースタッフのバージョン
    fn.start   = NULL;          // 前処理 start関数のポインタ
    fn.step    = NULL;          // simLoop関数のポインタ
    fn.command = NULL;
    fn.path_to_textures = "../../drawstuff/textures"; // テクスチャ
}

void setDrawStuff(dmFunctions *dmf)
{
    fn.version = DS_VERSION;    // ドロースタッフのバージョン
    fn.start   = dmf->start;    // 前処理 start関数のポインタ
    fn.step    = dmf->step;     // simLoop関数のポインタ
    fn.command = dmf->command;
    fn.path_to_textures = "../../drawstuff/textures"; // テクスチャ
}


/*** コールバック関数 ***/
void nearCallback(void *data, dGeomID o1, dGeomID o2)
{
    static const int N = 10; // 接触点数の最大値
    dContact contact[N];     // 接触点

    // 接触している物体のどちらかが地面ならisGroundに非0をセット
    int isGround = ((ground == o1) || (ground == o2));
    isGround = 1;

    // 2つのボディがジョイントで結合されていたら衝突検出しない
    dBodyID b1 = dGeomGetBody(o1);
    dBodyID b2 = dGeomGetBody(o2);
    if (b1 && b2 && dAreConnectedExcluding(b1,b2,dJointTypeContact)) return;

    // 衝突情報の生成 nは衝突点数
    int n = dCollide(o1,o2,N,&contact[0].geom,sizeof(dContact));
    if (isGround)
    {
        for (int i = 0; i < n; i++)
        {
            contact[i].surface.mode = dContactBounce; // 接触面の反発性を設定
            contact[i].surface.bounce = 0.1;          // 反発係数(0.0から1.0)
            contact[i].surface.bounce_vel = 0.05;      // 反発に必要な最低速度
            contact[i].surface.mu = 1.0; // dInfinity;        // 摩擦係数

            // 接触ジョイントの生成
            dJointID c = dJointCreateContact(world,contactgroup,
                                             &contact[i]);
            // 接触している２つの剛体を接触ジョイントにより拘束
            dJointAttach(c,dGeomGetBody(contact[i].geom.g1),
                         dGeomGetBody(contact[i].geom.g2));
        }
    }
}

/*** 力を物体の重心に加える　***/
void dmAddForce(dmObject *obj, double fx, double fy, double fz)
{
    dBodyAddForce(obj->body, fx, fy, fz);
}

/*** トルクを物体の重心に加える　***/
void dmAddTorque(dmObject *obj, double fx, double fy, double fz)
{
    dBodyAddTorque(obj->body, fx, fy, fz);
}

/*** シミュレーションの初期化 ***/
void dmInit()
{
    dInitODE();                              // ODEの初期化
    world = dWorldCreate();                  // 世界の創造
    dWorldSetGravity(world,0,0,-9.8);        // 重力設定

    space        = dHashSpaceCreate(0);   // 衝突用空間の創造
    contactgroup = dJointGroupCreate(0);  // ジョイントグループの生成
    ground = dCreatePlane(space,0,0,1,0); // 地面（平面ジオメトリ）の生成
}

/*** シミュレーションの終了 ***/
void dmClose()
{
    if (world != NULL) dWorldDestroy(world);  // 動力学世界の破壊
    if (space != NULL) dSpaceDestroy(space);  // 衝突検出空間の破壊
    dCloseODE();                             // ODEの終了
}

/*** 視点の設定　***/
void dmSetCamera(double x, double y, double z, double roll, double pitch, double yaw)
{
    float xyz[3], hpr[3];
    xyz[0] = (float) x;
    xyz[1] = (float) y;
    xyz[2] = (float) z;
    hpr[0] = (float) yaw;
    hpr[1] = (float) pitch;
    hpr[2] = (float) roll;
    dsSetViewpoint(xyz, hpr);
    dsSetSphereQuality(3);
}

void dmCreateSphere(dmObject *obj, double p[3], double R[12], double m, double r, double color[3])
{
    obj->body = dBodyCreate(world);          // ボールの生成
    obj->m = m;
    obj->r = r;

    obj->R = R;
    obj->p = p;
    obj->color = color;


    dMass mass;                              // 構造体massの宣言
    dMassSetZero(&mass);                     // 構造体massの初期化
    dMassSetSphereTotal(&mass,obj->m,obj->r);          // 構造体massに質量を設定
    dBodySetMass(obj->body,&mass);               // ボールにmassを設定

    dBodySetPosition(obj->body, obj->p[0], obj->p[1], obj->p[2]);  // ボールの位置(x,y,z)を設定
    dBodySetRotation(obj->body, obj->R);

    obj->geom = dCreateSphere(space,obj->r); // 球ジオメトリの生成
    dGeomSetBody(obj->geom, obj->body);      // ボディとジオメトリの関連付け
}

void dmCreateBox(dmObject *obj, double p[3], double R[12], double m, double side[3],  double color[3])
{
    obj->body = dBodyCreate(world);          // ボールの生成
    obj->m = m;

    obj->R = R;
    obj->p = p;
    obj->side = side;
    obj->color = color;

    dMass mass;                              // 構造体massの宣言
    dMassSetZero(&mass);                     // 構造体massの初期化
    dMassSetBoxTotal(&mass,obj->m,obj->side[0], obj->side[1], obj->side[2]);  // 構造体massに質量を設定
    dBodySetMass(obj->body,&mass);               // ボールにmassを設定

    dBodySetPosition(obj->body, obj->p[0], obj->p[1], obj->p[2]);  // ボールの位置(x,y,z)を設定
    dBodySetRotation(obj->body, obj->R);

    obj->geom = dCreateBox(space,obj->side[0], obj->side[1], obj->side[2]); // 球ジオメトリの生成
    dGeomSetBody(obj->geom, obj->body);      // ボディとジオメトリの関連付け
}

void dmCreateCylinder(dmObject *obj, double p[3], double R[12], double m, double r, double l, double color[3])
{
    obj->body = dBodyCreate(world);

    obj->m = m;
    obj->r = r;
    obj->l = l;

    obj->R = R;
    obj->p = p;
    obj->color = color;

    dMass mass;                              // 構造体massの宣言
    dMassSetZero(&mass);                     // 構造体massの初期化
    dMassSetBoxTotal(&mass,obj->m, 3, obj->r, obj->l);  // 構造体massに質量を設定
    dBodySetMass(obj->body,&mass);               // ボールにmassを設定

    dBodySetPosition(obj->body, obj->p[0], obj->p[1], obj->p[2]);  // ボールの位置(x,y,z)を設定
    dBodySetRotation(obj->body, obj->R);

    obj->geom = dCreateCylinder(space,obj->r, obj->l); // 円柱ジオメトリの生成
    dGeomSetBody(obj->geom, obj->body);      // ボディとジオメトリの関連付け
}

void dmCreateCapsule(dmObject *obj, double p[3], double R[12], double m, double r, double l, double color[3])
{
    obj->body = dBodyCreate(world);
    obj->m = m;
    obj->r = r;
    obj->l = l;

    obj->R = R;
    obj->p = p;
    obj->color = color;

    dMass mass;                              // 構造体massの宣言
    dMassSetZero(&mass);                     // 構造体massの初期化
    dMassSetBoxTotal(&mass,obj->m, 3, obj->r, obj->l);  // 構造体massに質量を設定
    dBodySetMass(obj->body,&mass);               // ボールにmassを設定

    dBodySetPosition(obj->body, obj->p[0], obj->p[1], obj->p[2]);  // ボールの位置(x,y,z)を設定
    dBodySetRotation(obj->body, obj->R);

    obj->geom = dCreateCapsule(space,obj->r, obj->l); // 円柱ジオメトリの生成
    dGeomSetBody(obj->geom, obj->body);      // ボディとジオメトリの関連付け
}


void dmSimStep()
{
    dSpaceCollide(space,0,&nearCallback); // 衝突検出関数
    dWorldStep(world,0.01);        // シミュレーションを1ステップ進める（精度が高い）
    dJointGroupEmpty(contactgroup); // ジョイントグループを空にする
}

void dmSimQuickStep()
{
    dSpaceCollide(space,0,&nearCallback); // 衝突検出関数
    dWorldQuickStep(world,0.01);     // シミュレーションを1ステップ進める（高速）
    dJointGroupEmpty(contactgroup); // ジョイントグループを空にする
}

void dmDraw(dmObject *obj) /*** 物体の描画 ***/
{
    if (!obj->geom) return;
    obj->p = (double *) dGeomGetPosition(obj->geom);
    obj->R = (double *) dGeomGetRotation(obj->geom);

    dsSetColor(obj->color[0],obj->color[1],obj->color[2]);  // 色の設定(r,g,b)
    // printf("color=%.1f %.1f %.1f\n",obj->color[0],obj->color[1],obj->color[2]);

    int type = dGeomGetClass(obj->geom);

    switch (type)
    {
    case dBoxClass:
    {
        dVector3 sides;
        dGeomBoxGetLengths(obj->geom,sides);
        dsDrawBox(obj->p,obj->R,sides);
    }
    break;

    case dSphereClass:
        dsDrawSphere(obj->p,obj->R,dGeomSphereGetRadius(obj->geom));
        break;

    case  dCapsuleClass:
    {
        dReal radius,length;
        dGeomCapsuleGetParams(obj->geom,&radius,&length);
        dsDrawCapsule(obj->p,obj->R,length,radius);
    }
    break;

    case dCylinderClass:
    {
        dReal radius,length;
        dGeomCylinderGetParams(obj->geom,&radius,&length);
        dsDrawCylinder(obj->p,obj->R,length,radius);
    }
    break;

    default:
        printf("Bad geometry type \n");
    }
}



dJointID dmJointCreateFixed()
{
    return dJointCreateFixed(world, 0);
}

dJointID dmJointCreateSlider()
{
    return dJointCreateSlider(world, 0);
}


dJointID dmJointCreateHinge()
{
    return dJointCreateHinge(world, 0);
}

void dmLoop(int w, int h, dmFunctions *dmf)
{
    setDrawStuff(dmf);
    dsSimulationLoop(0,0,w,h,&fn);
}

void dmLoop()
{
    setDrawStuff();
    dsSimulationLoop(0,0,320,240,&fn);
}

void dmLoop(int w, int h)
{
    setDrawStuff();
    dsSimulationLoop(0,0,w,h,&fn);
}

static void makeBall()
{
    dMass m1;

    dReal  ball_mass   = 0.45;
    ball.body    = dBodyCreate(world);

    dMassSetZero(&m1);
    dMassSetSphereTotal(&m1,ball_mass,BALL_R);
    dMassAdjust(&m1,ball_mass);
    dBodySetMass(ball.body,&m1);

    ball.geom = dCreateSphere(space,BALL_R);
    dGeomSetBody(ball.geom,ball.body);
    dBodySetPosition(ball.body,BALL_P[0],BALL_P[1],BALL_P[2]);
}

void makeRobot()
{
    dMatrix3 R;
    dMass mass, mass2, mass3;

    // 台車
    base.body  = dBodyCreate(world);
    dMassSetZero(&mass);
    dMassSetBoxTotal(&mass,BASE_M,BASE_S[0],BASE_S[1],BASE_S[2]);
    dBodySetMass(base.body,&mass);

    base.geom = dCreateBox(space,BASE_S[0],BASE_S[1],BASE_S[2]);
    dGeomSetBody(base.geom,base.body);
    dBodySetPosition(base.body,START_X,START_Y,START_Z);

    // 首
    neck.body  = dBodyCreate(world);
    dMassSetZero(&mass2);
    dMassSetCapsuleTotal(&mass2,NECK_M,3,NECK_R,NECK_L); // 引数3はカプセルの長軸方向がz軸
    dMassAdjust(&mass2,NECK_M);
    dBodySetMass(neck.body,&mass2);

    neck.geom = dCreateCCylinder(space,NECK_R,NECK_L);
    dGeomSetBody(neck.geom,neck.body);
    dBodySetPosition(neck.body,START_X, START_Y+0.5*BASE_S[1]-NECK_R,
                     START_Z+0.5*NECK_L);
    dGeomDisable(neck.geom);

    // カメラ
    camera.body  = dBodyCreate(world);
    dMassSetZero(&mass3);
    dMassSetBoxTotal(&mass3,CAMERA_M,CAMERA_S[0],CAMERA_S[1],CAMERA_S[2]);
    dMassAdjust(&mass3,CAMERA_M);
    dBodySetMass(camera.body,&mass3);

    camera.geom = dCreateBox(space,CAMERA_S[0],CAMERA_S[1],CAMERA_S[2]);
    dGeomSetBody(camera.geom,camera.body);
    dBodySetPosition(camera.body,START_X, START_Y+0.5*BASE_S[1]-0.1*CAMERA_S[1],
                     START_Z+NECK_L);
    dGeomDisable(camera.geom);


    // 車輪
    dRFromAxisAndAngle(R,0,1,0,M_PI/2.0);
    for (int i = 0; i < NUM; i++)
    {
        wheel[i].body = dBodyCreate(world);

        dMassSetZero(&mass);
        if ((i % 2) == 0)
        {
            dMassSetCylinderTotal(&mass,WH_M0, 2, WH_R0, WH_W);
            dBodySetMass(wheel[i].body,&mass);
            wheel[i].geom = dCreateCylinder(space, WH_R0, WH_W);
        }
        else
        {
            dMassSetCylinderTotal(&mass,WH_M1, 2, WH_R1, WH_W);
            dBodySetMass(wheel[i].body,&mass);
            wheel[i].geom = dCreateCylinder(space, WH_R1, WH_W);
        }
        dGeomSetBody(wheel[i].geom, wheel[i].body);
        dBodySetRotation(wheel[i].body,R);
    }

    dReal w1x = 0.5 * (BASE_S[1]+WH_W);
    dReal w1z = START_Z - 0.5 * BASE_S[2];
    dReal w0y = 0.5 * BASE_S[1] - WH_R0;
    dReal w0z = START_Z - 0.5 * BASE_S[2] - WH_R0;

    dBodySetPosition(wheel[0].body,    0,  w0y, w0z);
    dBodySetPosition(wheel[1].body, -w1x,    0, w1z);
    dBodySetPosition(wheel[2].body,    0, -w0y, w0z);
    dBodySetPosition(wheel[3].body,  w1x,    0, w1z);

    // ヒンジジョイント
    for (int i = 0; i < NUM; i++)
    {
        joint[i] = dJointCreateHinge(world,0);
        dJointAttach(joint[i], base.body, wheel[i].body);
    }
    dJointSetHingeAxis(joint[0],1, 0, 0);
    dJointSetHingeAxis(joint[1],1, 0, 0);
    dJointSetHingeAxis(joint[2],1, 0, 0);
    dJointSetHingeAxis(joint[3],1, 0, 0);
    dJointSetHingeAnchor(joint[0],    0,  w0y, w0z);
    dJointSetHingeAnchor(joint[1], -w1x,    0, w1z);
    dJointSetHingeAnchor(joint[2],    0, -w0y, w0z);
    dJointSetHingeAnchor(joint[3],  w1x,    0, w1z);

// joints
    neck_joint = dJointCreateHinge(world,0);
    dJointAttach(neck_joint,neck.body,base.body);
    dJointSetHingeAxis(neck_joint,0, 0, 1);
    dJointSetHingeAnchor(neck_joint, START_X, START_Y+0.5*BASE_S[1]-NECK_R,
                         START_Z);
    dJointSetHingeParam(neck_joint,dParamLoStop, 0);
    dJointSetHingeParam(neck_joint,dParamHiStop, 0);


    // camera joints
    camera_joint = dJointCreateHinge(world,0);
    dJointAttach(camera_joint,neck.body,camera.body);
    dJointSetHingeAxis(camera_joint,1,0,0);
    dJointSetHingeAnchor(camera_joint,START_X, START_Y+0.5*BASE_S[1]-0.1*CAMERA_S[1],
                         START_Z+NECK_L);
    dJointSetHingeParam(camera_joint,dParamLoStop, 0.05 * M_PI);
    dJointSetHingeParam(camera_joint,dParamHiStop, 0.05 * M_PI);


}

static void control()
{
    double fMax = 100;         // 最大発揮トルク[Nm]

    dJointSetHingeParam(joint[1],  dParamVel, v_l );
    dJointSetHingeParam(joint[1], dParamFMax, fMax);
    dJointSetHingeParam(joint[3],  dParamVel, v_r );
    dJointSetHingeParam(joint[3], dParamFMax, fMax);
}

void drawBall()
{
    dsSetColor(1.0,0.0,0.0);
    dsDrawSphere(dGeomGetPosition(ball.geom),
                 dGeomGetRotation(ball.geom),BALL_R);
}

void drawRobot()
{
    // 台車
    //dsSetColor(1.3,1.3,0.0);
    dsSetColor(1.1,1.1,1.1);
    dsDrawBox(dGeomGetPosition(base.geom),
              dGeomGetRotation(base.geom),BASE_S);

    // 首
    dsDrawCylinder(dBodyGetPosition(neck.body),
                   dBodyGetRotation(neck.body),NECK_L,NECK_R);

    // カメラ
    dsDrawBox(dGeomGetPosition(camera.geom),
              dGeomGetRotation(camera.geom),CAMERA_S);

    // 車輪
    //dsSetColor(1.1,1.1,1.1);
    dsSetColor(0.1, 0.1, 0.1);
    for (int i=0; i< NUM; i++)
    {
        if ((i % 2) == 0)
        {
            dsDrawCylinder(dBodyGetPosition(wheel[i].body),
                           dBodyGetRotation(wheel[i].body),WH_W,WH_R0);
        }
        else
        {
            dsDrawCylinder(dBodyGetPosition(wheel[i].body),
                           dBodyGetRotation(wheel[i].body),WH_W,WH_R1);
        }
    }
}

// フィールド座標系からロボット座標系への変換  プログラム4.2
static void field2RobotPos(const dReal p[2], const dReal c[2], dReal theta,
                           dReal pr[2], dReal *r, dReal *phi)
{
    pr[0] =   (p[0]-c[0]) * cos(theta) + (p[1]-c[1]) * sin(theta);
    pr[1] = - (p[0]-c[0]) * sin(theta) + (p[1]-c[1]) * cos(theta);

    *r   = sqrt(pr[0] * pr[0] + pr[1] * pr[1]);  // 物体までの距離
    *phi = atan2(pr[0],pr[1]);                   // 物体の角度
}

/*** 車輪の制御 ***/
void motor(int right, int left)
{
    double fMax = 100;  // 最大トルク

    if (right >=  100) right =  100;
    if (right <= -100) right = -100;
    if (left  >=  100) left  =  100;
    if (left  <= -100) left  = -100;

    double omega_r = right/ (100.0 * WH_R1);
    double omega_l = left / (100.0 * WH_R1);


    dJointSetHingeParam(joint[1],  dParamVel, omega_l);
    dJointSetHingeParam(joint[1], dParamFMax, fMax);
    dJointSetHingeParam(joint[3],  dParamVel, omega_r);
    dJointSetHingeParam(joint[3], dParamFMax, fMax);
}

/*** ロボットのビジョン関数 ***/
double vision(dReal *r, dReal *phi, dBodyID obj_body)
{
    const dReal *pa,*pb,*pc;
    dReal pr[2], view_angle = M_PI/36.0;  // 相対位置，視野角

    pa  = dBodyGetPosition(obj_body);    // 物体の絶対座標
    pb  = dBodyGetPosition(base.body);   // 台車の絶対座標
    pc  = dBodyGetPosition(camera.body); // カメラの絶対座標

    dReal theta  = atan2(pc[0]-pb[0],pc[1]-pb[1]); // ロボットの針路
    // 絶対座標（右ねじ）と相対座標（左ねじ）で角度の取り方が逆
    theta = - theta;

    field2RobotPos(pa, pb, theta, pr, r, phi);

    if ((-view_angle <= *phi) && (*phi <= view_angle))
    {
        return (double) *r;
    }
    else
    {
        return 0;
    }
}

/*** ロボットのビジョン関数 ***/
int ballSensor()
{
    dReal r, phi;
    const dReal *pa,*pb,*pc;
    dReal pr[2], view_angle = M_PI/36.0;  // 相対位置，視野角

    pa  = dBodyGetPosition(ball.body);    // 物体の絶対座標
    pb  = dBodyGetPosition(base.body);   // 台車の絶対座標
    pc  = dBodyGetPosition(camera.body); // カメラの絶対座標

    dReal theta  = atan2(pc[0]-pb[0],pc[1]-pb[1]); // ロボットの針路
    // 絶対座標（右ねじ）と相対座標（左ねじ）で角度の取り方が逆
    theta = - theta;

    field2RobotPos(pa, pb, theta, pr, &r, &phi);

    if ((-view_angle <= phi) && (phi <= view_angle))
    {
        return (int) (2000.0/(100.0*r));
    }
    else
    {
        return  rand()%7;
    }
}



/***   その場回転  ***/
void turn(int power)
{
    motor(power, -power);
}

/*** 追跡  phi:物体の方位, speed:車輪の回転速度 ***/
void track(dReal phi, int power)
{
    dReal k = 0.5;
    if (phi > 0) motor(    power,  k * power);
    else         motor(k * power,      power);
}

/*** カメラの位置と姿勢設定 ***/
void setCamera()
{
    float x = 3.0, y = 0.0, z = 1.0;    // カメラの位置
    float roll = 0, pitch = 0, yaw = -180; // カメラの方向[°]
    dmSetCamera(x,y,z,roll,pitch,yaw);  // カメラの設定
}


/*** 描画用構造体の設定　***/
void setDraw()
{
    dmf.start   = &setCamera;
    dmf.step    = &simLoop;
    dmf.command = NULL; // &command;
}

void simLoop(int pause)
{
    CN1 = ballSensor();

    user_main();

    dmSimStep();
    drawRobot();
    drawBall();
}

int main()         /*** main関数 ***/
{
    dmInit(); // ODEの初期化
    setDraw(); // 描画用構造体の設定
    dWorldSetERP(world, 0.8);
    dWorldSetCFM(world, 1e-4);
    makeBall();    // ボールの生成
    makeRobot();   // ロボットの生成
    dmLoop(800, 600, &dmf);  // シミュレーションループ ウインドウの幅，高
    dmClose();          // ODEの終了
    return 0;
}



